// C Standard Library
#include <assert.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <mpi.h>
// fmpi
#include <fmpi.h>
// Internal
#include "as.h"

#define T int64_t

FMPI_TASK_FUTHARK(array_sum, 1)

int main(int argc, char * argv[])
{
    struct fmpi_ctx * ctx = fmpi_init(&argc, &argv);
    if(ctx == NULL) {
        fprintf(stderr, "fmpi_init() failed!\n");
        return EXIT_FAILURE;
    }
    T in[] = {
        1,0,0,0,0,0,0,0,
        0,0,0,0,0,0,0,1,
        2,0,0,0,0,0,0,0,
        0,0,0,0,0,0,0,2,
        3,0,0,0,0,0,0,0,
        0,0,0,0,0,0,0,3,
        4,0,0,0,0,0,0,0,
        0,0,0,0,0,0,0,4
    };
    const size_t in_size = sizeof(in)/sizeof(T);
    T out = -1;
    struct fmpi_task array_sum_task = FMPI_REGISTER_SYNC_TASK(
        ctx, array_sum, fmpi_no_stencil(),
        fmpi_data_out(ctx, &out),
        fmpi_data_1d_in(ctx, in, in_size)
    );
    if(fmpi_run_task(ctx, &array_sum_task) != FMPI_SUCCESS) {
        printf("Error rank=%d: fmpi_run_task()\n", fmpi_world_rank(ctx));
    }
    printf("rank=%d local sum=%ld\n", fmpi_world_rank(ctx), out);
    if(fmpi_task_finalize(ctx, &array_sum_task, FMPI_TASK_OP_SUM) != FMPI_SUCCESS) {
        printf("Error rank=%d: fmpi_task_finalize()\n", fmpi_world_rank(ctx));
    }
    fmpi_root_printf(ctx, "global sum=%ld\n", out);
    fmpi_exit(&ctx);
    return EXIT_SUCCESS;
}
