// C Standard Library
#include <inttypes.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
// fmpi
#include <fmpi.h>
// Internal
#include "gol.h"

#define T u64
#define STEP_CNT 20
#define X_LEN 8
#define Y_LEN 8

FMPI_TASK_FUTHARK(gol, 1)

int main(int argc, char * argv[])
{
    struct fmpi_ctx * ctx = fmpi_init(&argc, &argv);
    if(ctx == NULL) {
        fprintf(stderr, "fmpi_init() failed!\n");
        return EXIT_FAILURE;
    }
    T in[] = {
        0,1,0,0,0,0,0,0,
        0,0,1,0,0,0,0,0,
        1,1,1,0,0,0,0,0,
        0,0,0,0,0,0,0,0,
        0,0,0,0,0,0,0,0,
        0,0,0,0,0,0,0,0,
        0,0,0,0,0,0,0,0,
        0,0,0,0,0,0,0,0,
    };
    #define in_size (sizeof(in)/sizeof(T))
    T out[in_size] = {0};
    struct fmpi_task gol_task = FMPI_REGISTER_SYNC_TASK(
        ctx, gol, fmpi_stencil_square(1),
        fmpi_data_2d_out(ctx, out, X_LEN, Y_LEN),
        fmpi_data_2d_in(ctx, in, X_LEN, Y_LEN)
    );
    for(size_t i = 0; i < STEP_CNT; i++) {
        if(fmpi_run_task(ctx, &gol_task) != FMPI_SUCCESS) {
            printf("Error rank=%d: fmpi_run_task()\n", fmpi_world_rank(ctx));
        }
    }
    if(fmpi_task_finalize(ctx, &gol_task, FMPI_TASK_OP_GATHER) != FMPI_SUCCESS) {
        printf("Error rank=%d: fmpi_task_finalize()\n", fmpi_world_rank(ctx));
    }
    if(fmpi_is_root(ctx)) {
        T expected[] = {
            0,0,0,0,0,0,0,0,
            0,0,0,0,0,0,0,0,
            0,0,0,0,0,0,0,0,
            0,0,0,0,0,0,0,0,
            0,0,0,0,0,0,0,0,
            0,0,0,0,0,0,1,0,
            0,0,0,0,0,0,0,1,
            0,0,0,0,0,1,1,1,
        };
        printf("Expected =>\n");
        for(size_t i = 0; i < in_size; i++) {
            printf("%"PRIu64, expected[i]);
            if((i+1) % X_LEN == 0) {
                printf("\n");
            }
        }
        printf("Actual =>\n");
        size_t correct_cnt = 0;
        for(size_t i = 0; i < in_size; i++) {
            if(out[i] == expected[i]) {
                correct_cnt++;
            }
            printf("%"PRIu64, out[i]);
            if((i+1) % X_LEN == 0) {
                printf("\n");
            }
        }
        if(correct_cnt == in_size) {
            printf("OK!\n");
        }
    }
    fmpi_exit(&ctx);
    return EXIT_SUCCESS;
}
