// C Standard Library
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
// fmpi
#include <fmpi.h>
// Internal
#include "r110.h"

#define T u8
#define STEP_CNT 20

FMPI_TASK_FUTHARK(rule_110, 1)

int main(int argc, char * argv[])
{
    struct fmpi_ctx * ctx = fmpi_init(&argc, &argv);
    if(ctx == NULL) {
        fprintf(stderr, "fmpi_init() failed!\n");
        return EXIT_FAILURE;
    }
    T in[] = {
        1,1,0,1,0,0,0,0,1,0,0,1,0,0,0,1,0,0,0,0,0,1
    };
    #define in_size (sizeof(in)/sizeof(T))
    T out[in_size];
    struct fmpi_task rule_110_task = FMPI_REGISTER_SYNC_TASK(
        ctx, rule_110, fmpi_stencil_1d(1),
        fmpi_data_1d_out(ctx, out, in_size),
        fmpi_data_1d_in(ctx, in, in_size)
    );
    for(size_t i = 0; i < STEP_CNT; i++) {
        if(fmpi_run_task(ctx, &rule_110_task) != FMPI_SUCCESS) {
            printf("Error rank=%d: fmpi_run_task()\n", fmpi_world_rank(ctx));
        }
    }
    if(fmpi_task_finalize(ctx, &rule_110_task, FMPI_TASK_OP_GATHER) != FMPI_SUCCESS) {
        printf("Error rank=%d: fmpi_task_finalize()\n", fmpi_world_rank(ctx));
    }
    if(fmpi_is_root(ctx)) {
        T result[] = {1,1,1,0,1,1,1,0,0,1,1,0,0,1,0,0,1,1,0,1,1,1};
        size_t correct_cnt = 0;
        printf("Expected: 1110111001100100110111\n");
        printf("Actual  : ");
        for(size_t i = 0; i < in_size; i++) {
            if(out[i] == result[i]) {
                correct_cnt++;
            }
            printf("%hhd", out[i]);
        }
        printf("\n");
        if(correct_cnt == in_size) {
            printf("OK!\n");
        }
    }
    fmpi_exit(&ctx);
    return EXIT_SUCCESS;
}
