// SPDX-License-Identifier: 0BSD
/*!
 * @file
 * @license{
 * BSD Zero Clause License
 *
 * Copyright (c) 2022 by Raphael Bach
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
 * REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
 * INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
 * OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 * }
 */
/*==============================================================================
    GUARD
==============================================================================*/
#ifndef FMPI_DATA_H_20211228231216
#define FMPI_DATA_H_20211228231216
/*==============================================================================
    INCLUDE
==============================================================================*/
// C Standard Library
#include <stddef.h> // size_t
// Internal
#include "internal/fmpi_common.h"
#include "internal/fmpi_type.h"
#include "internal/generic/fmpi_data_generic.h"
/*==============================================================================
    TYPE
==============================================================================*/
/*------------------------------------------------------------------------------
    fmpi_data
------------------------------------------------------------------------------*/
/**
 * Holds informations about data such as its type, count, size, etc.
 */
typedef struct fmpi_data {
    struct fmpi_type type;        //!< Type of the data.
    size_t cnt;                   //!< Total number of elements.
    size_t size;                  //!< Total size `(cnt * type.size)`.
    size_t dim_len[FMPI_DIM_MAX]; //!< Length of each dimension.
    size_t dim_cnt;               //!< Number of dimensions.
    void * raw;                   //!< Pointer to user data.
    void * gpu;                   //!< Pointer to data returned by the GPU.
} fmpi_data;
/*==============================================================================
    PUBLIC FUNCTION
==============================================================================*/
/*------------------------------------------------------------------------------
    fmpi_data_out()
------------------------------------------------------------------------------*/
#define fmpi_data_out(ctx, data) \
    FMPI_DATA_FUNC(out, data)(ctx, data)
/*------------------------------------------------------------------------------
    fmpi_data_1d()
------------------------------------------------------------------------------*/
/**
 * TODO
 */
#define fmpi_data_1d_in(ctx, data, x) \
    FMPI_DATA_FUNC(1d_in, data)(ctx, data, x, 1, 1)

#define fmpi_data_1d_out(ctx, data, x) \
    FMPI_DATA_FUNC(1d_out, data)(ctx, data, x, 1, 1)
/*------------------------------------------------------------------------------
    fmpi_data_2d()
------------------------------------------------------------------------------*/
/**
 * TODO
 */
#define fmpi_data_2d_in(ctx, data, x, y) \
    FMPI_DATA_FUNC(2d_in, data)(ctx, data, x, y, 1)

#define fmpi_data_2d_out(ctx, data, x, y) \
    FMPI_DATA_FUNC(2d_out, data)(ctx, data, x, y, 1)
/*------------------------------------------------------------------------------
    fmpi_data_3d()
------------------------------------------------------------------------------*/
/**
 * TODO
 */
#define fmpi_data_3d_in(ctx, data, x, y, z) \
    FMPI_DATA_FUNC(3d_in, data)(ctx, data, x, y, z)

#define fmpi_data_3d_out(ctx, data, x, y, z) \
    FMPI_DATA_FUNC(3d_out, data)(ctx, data, x, y, z)
/*==============================================================================
    GUARD
==============================================================================*/
#endif // FMPI_DATA_H_20211228231216
