// SPDX-License-Identifier: 0BSD
/*!
 * @file
 * @license{
 * BSD Zero Clause License
 *
 * Copyright (c) 2022 by Raphael Bach
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
 * REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
 * INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
 * OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 * }
 */
/*==============================================================================
    GUARD
==============================================================================*/
#ifndef FMPI_DOMAIN_H_20220520194220
#define FMPI_DOMAIN_H_20220520194220
/*==============================================================================
    INCLUDE
==============================================================================*/
// C Standard Library
#include <stddef.h> // size_t
// Internal
#include "fmpi_data.h"
#include "fmpi_stencil.h"
/*==============================================================================
    TYPE
==============================================================================*/
// Forward declaration from `fmpi_ctx.h`
struct fmpi_ctx;
/*------------------------------------------------------------------------------
    fmpi_domain
------------------------------------------------------------------------------*/
/**
 * TODO
 */
typedef struct fmpi_domain {
    const struct fmpi_data * data; //!< Data composing the domain.
    struct fmpi_data inner;
    struct fmpi_data halo;
    struct fmpi_stencil stencil;
    size_t part_cnt;          //!< Number of partitions.
} fmpi_domain;
/*==============================================================================
    PUBLIC FUNCTION
==============================================================================*/
/*------------------------------------------------------------------------------
    fmpi_new_domain()
------------------------------------------------------------------------------*/
struct fmpi_domain fmpi_new_domain(
    const struct fmpi_ctx * ctx, const struct fmpi_data * data,
    struct fmpi_stencil stencil
);
/*------------------------------------------------------------------------------
    fmpi_domain_set_inner()
------------------------------------------------------------------------------*/
void fmpi_domain_set_inner(
    const struct fmpi_ctx * ctx, const struct fmpi_domain * domain,
    const void * data
);
/*------------------------------------------------------------------------------
    fmpi_halo_exchange_1d()
------------------------------------------------------------------------------*/
int fmpi_halo_exchange_1d(
    const struct fmpi_ctx * ctx, const struct fmpi_domain * domain
);
/*------------------------------------------------------------------------------
    fmpi_halo_exchange_2d()
------------------------------------------------------------------------------*/
int fmpi_halo_exchange_2d(
    const struct fmpi_ctx * ctx, const struct fmpi_domain * domain
);
/*==============================================================================
    GUARD
==============================================================================*/
#endif // FMPI_DOMAIN_H_20220520194220
