// SPDX-License-Identifier: 0BSD
/*!
 * @file
 * @license{
 * BSD Zero Clause License
 *
 * Copyright (c) 2022 by Raphael Bach
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
 * REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
 * INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
 * OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 * }
 */
/*==============================================================================
    GUARD
==============================================================================*/
#ifndef FMPI_REDUCE_H_20211218002000
#define FMPI_REDUCE_H_20211218002000
/*==============================================================================
    INCLUDE
==============================================================================*/
// Internal
#include "internal/generic/fmpi_reduce_generic.h"
/*==============================================================================
    PUBLIC FUNCTION
==============================================================================*/
/*------------------------------------------------------------------------------
    fmpi_local_reduce_prod()
------------------------------------------------------------------------------*/
/**
 * Return the locally computed product of all elements in an \p{array}.
 *
 * @param[in,out] ctx    : Pointer to a fmpi_ctx structure.
 * @param[in]     array  : Array to reduce.
 * @param[in]     length : Length of the \p{array}.
 *
 * @return The product of all elements in \p{array}.
 *
 * @warning
 * - \p{ctx} must be a valid pointer allocated with fmpi_init().
 * - \p{array} must not be `NULL`.
 */
#define fmpi_local_reduce_prod(ctx, array, length) \
    FMPI_GENERIC_FUNC(array, local_reduce_prod, FMPI_REDUCE_PROD_TYPES)(ctx, array, length)
/*------------------------------------------------------------------------------
    fmpi_reduce_prod()
------------------------------------------------------------------------------*/
/**
 * TODO
 *
 * @param ctx : TODO
 * @param array : TODO
 */
#define fmpi_reduce_prod(ctx, array) \
    FMPI_GENERIC_FUNC(array, reduce_prod, FMPI_REDUCE_PROD_TYPES)(ctx, array)
/*==============================================================================
    GUARD
==============================================================================*/
#endif // FMPI_REDUCE_H_20211218002000
