// SPDX-License-Identifier: 0BSD
/*!
 * @file
 * @license{
 * BSD Zero Clause License
 *
 * Copyright (c) 2022 by Raphael Bach
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
 * REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
 * INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
 * OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 * }
 */
/*==============================================================================
    GUARD
==============================================================================*/
#ifndef FMPI_TASK_H_20220513032517
#define FMPI_TASK_H_20220513032517
/*==============================================================================
    INCLUDE
==============================================================================*/
// C Standard Library
#include <stddef.h> // size_t
// Internal
#include "fmpi_data.h"
#include "fmpi_domain.h"
#include "fmpi_stencil.h"
#include "internal/fmpi_futhark.h"
#include "internal/generic/fmpi_task_generic.h"
/*==============================================================================
    DEFINE
==============================================================================*/
/**
 * Maximum number of arguments that can be passed to a task.
 * @todo Find a meaningful value.
 */
#define FMPI_TASK_ARGS_MAX 32
/**
 * Maximum number of tasks that can be created.
 * @todo Find a meaningful value.
 */
#define FMPI_TASK_MAX 32
/*==============================================================================
    TYPE
==============================================================================*/
/*------------------------------------------------------------------------------
    fmpi_task_op
------------------------------------------------------------------------------*/
typedef enum fmpi_task_op {
    FMPI_TASK_OP_NONE = 0,
    FMPI_TASK_OP_SUM,
    FMPI_TASK_OP_GATHER
} fmpi_task_op;
/*------------------------------------------------------------------------------
    fmpi_task_args
------------------------------------------------------------------------------*/
/**
 * TODO
 */
typedef struct fmpi_task_args {
    struct fmpi_data in[FMPI_TASK_ARGS_MAX]; //!< TODO
    struct fmpi_data out;                        //!< TODO
    size_t cnt;                                  //!< TODO
} fmpi_task_args;
/*------------------------------------------------------------------------------
    fmpi_task_func
------------------------------------------------------------------------------*/
/**
 * TODO
 *
 * @param[in] ctx  : A pointer to a fmpi_ctx.
 * @param[in] args : A pointer to a fmpi_task_args.
 *
 * @return Nothing.
 *
 * @warning
 * - \b [UB] \p{ctx} must not be `NULL`.
 * - \b [UB] \p{args} must not be `NULL`.
 *
 * @example{
 *  TODO
 * }
 */
typedef int (* fmpi_task_func)(
    const struct fmpi_ctx * ctx, const struct fmpi_task_args * args
);
/*------------------------------------------------------------------------------
    fmpi_task_type
------------------------------------------------------------------------------*/
typedef enum fmpi_task_type {
    FMPI_TASK_TYPE_SYNC,
    FMPI_TASK_TYPE_ASYNC
} fmpi_task_type;
/*------------------------------------------------------------------------------
    fmpi_task
------------------------------------------------------------------------------*/
/**
 * TODO
 */
typedef struct fmpi_task {
    fmpi_task_func func;        //!< TODO
    const char * name; //!< Name of the function.
    enum fmpi_task_type type; //!< Sync/async
    struct fmpi_task_args args; //!< TODO
    struct fmpi_domain domains[FMPI_TASK_ARGS_MAX]; //!< TODO
    struct fmpi_stencil stencil; //!< TODO
} fmpi_task;
/*==============================================================================
    PUBLIC FUNCTION
==============================================================================*/
/*------------------------------------------------------------------------------
    fmpi_task_register()
------------------------------------------------------------------------------*/
/**
 * TODO
 *
 * @param[in] ctx  : A pointer to a fmpi_ctx.
 * @param[in] func : A function with a fmpi_task_func signature.
 * @param[in] args : A pointer to a fmpi_task_args.
 *
 * @return A fmpi_task structure.
 *
 * @warning
 * - \b [UB] \p{ctx} must not be `NULL`.
 * - \b [UB] \p{func} must not be `NULL`.
 * - \b [UB] \p{args} must not be `NULL`.
 *
 * @example{
 *  TODO
 * }
 */
struct fmpi_task fmpi_task_register(
    struct fmpi_ctx * ctx, fmpi_task_func func, const char * name,
    enum fmpi_task_type type, struct fmpi_stencil stencil,
    const struct fmpi_task_args * args
);
/*------------------------------------------------------------------------------
    fmpi_task_run_sync()
------------------------------------------------------------------------------*/
/**
 * Runs a task synchronously.
 *
 * @param[in] ctx  : A pointer to a fmpi_ctx.
 * @param[in] task : A pointer to a fmpi_task.
 *
 * @return
 * - @success: `0`.
 * - @failure: An error code different from `0`.
 *
 * @warning
 * - \b [UB] \p{ctx} must not be `NULL`.
 * - \b [UB] \p{task} must not be `NULL`.
 *
 * @example{
 *  TODO
 * }
 */
int fmpi_task_run_sync(const struct fmpi_ctx * ctx, struct fmpi_task * task);
/*------------------------------------------------------------------------------
    fmpi_task_run_async()
------------------------------------------------------------------------------*/
/**
 * Runs a task asynchronously.
 *
 * @param[in] ctx  : A pointer to a fmpi_ctx.
 * @param[in] task : A pointer to a fmpi_task.
 *
 * @return
 * - @success: `0`.
 * - @failure: An error code different from `0`.
 *
 * @warning
 * - \b [UB] \p{ctx} must not be `NULL`.
 * - \b [UB] \p{task} must not be `NULL`.
 *
 * @example{
 *  TODO
 * }
 */
int fmpi_task_run_async(const struct fmpi_ctx * ctx, const struct fmpi_task * task);
/*------------------------------------------------------------------------------
    fmpi_task_finalize()
------------------------------------------------------------------------------*/
int fmpi_task_finalize(
    const struct fmpi_ctx * ctx, const struct fmpi_task * task,
    enum fmpi_task_op op
);
/*==============================================================================
    MACRO
==============================================================================*/
/*------------------------------------------------------------------------------
    FMPI_TASK_FUTHARK()
------------------------------------------------------------------------------*/
/**
 * Declares and defines a task running a `futhark` function.
 *
 * @param[in] func    : TODO
 * @param[in] arg_cnt : TODO
 *
 * @return Nothing.
 *
 * @warning
 * - \b [UB] \p{func} must not be `NULL`.
 *
 * @example{
 *  FMPI_TASK_FUTHARK(array_sum, 1)
 * }
 */
#define FMPI_TASK_FUTHARK(func, arg_cnt) \
    FMPI_TASK_DEFINITION(func, arg_cnt)
/*==============================================================================
    GUARD
==============================================================================*/
#endif // FMPI_TASK_H_20220513032517
