// SPDX-License-Identifier: 0BSD
/*!
 * @file
 * @license{
 * BSD Zero Clause License
 *
 * Copyright (c) 2022 by Raphael Bach
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
 * REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
 * INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
 * OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 * }
 */
/*==============================================================================
    GUARD
==============================================================================*/
#ifndef FMPI_FUTHARK_H_20211219234207
#define FMPI_FUTHARK_H_20211219234207
/*==============================================================================
    INCLUDE
==============================================================================*/
// C Standard Library
#include <stddef.h> // size_t
// Internal
#include "fmpi_error.h"
#include "internal/fmpi_type.h"
#include "internal/generic/fmpi_futhark_generic.h"
/*==============================================================================
    STRUCT
==============================================================================*/
/*------------------------------------------------------------------------------
    fmpi_futhark_ctx()
------------------------------------------------------------------------------*/
/**
 * TODO
 */
typedef struct fmpi_futhark_ctx {
    struct futhark_context * ctx;          //!< TODO
    struct futhark_context_config * cfg;   //!< TODO
    struct fmpi_error_handler err_handler; //!< TODO
} fmpi_futhark_ctx;
/*==============================================================================
    PUBLIC FUNCTION
==============================================================================*/
/*------------------------------------------------------------------------------
    fmpi_futhark_init()
------------------------------------------------------------------------------*/
/**
 * TODO
 *
 * @param[in] err_handler : TODO
 *
 * @return
 * - @success: TODO
 * - @failure: TODO
 *
 * @warning TODO
 *
 * @example{
 *  TODO
 * }
 */
struct fmpi_futhark_ctx * fmpi_futhark_init(struct fmpi_error_handler err_handler);
/*------------------------------------------------------------------------------
    fmpi_futhark_exit()
------------------------------------------------------------------------------*/
/**
 * TODO
 *
 * @param[in] ctx : TODO
 *
 * @return
 * - @success: TODO
 * - @failure: TODO
 *
 * @warning TODO
 *
 * @example{
 *  TODO
 * }
 */
int fmpi_futhark_exit(struct fmpi_futhark_ctx ** ctx);
/*------------------------------------------------------------------------------
    fmpi_futhark_sync()
------------------------------------------------------------------------------*/
/**
 * TODO
 *
 * @param[in] ctx : TODO
 *
 * @return
 * - @success: TODO
 * - @failure: TODO
 *
 * @warning TODO
 *
 * @example{
 *  TODO
 * }
 */
int fmpi_futhark_sync(const struct fmpi_futhark_ctx * ctx);
/*------------------------------------------------------------------------------
    fmpi_futhark_has_error()
------------------------------------------------------------------------------*/
/**
 * Check if the last function called before calling
 * fmpi_has_futhark_error() generated an error.
 *
 * @param[in,out] ctx  : Pointer to a fmpi_ctx structure.
 * @param[in]     func : Function name.
 *
 * @return
 * - `true` if an error is detected.
 * - `false` if no error was detected.
 *
 * @warning
 * - \p{ctx} must be a valid pointer allocated with fmpi_init().
 * - \p{func} must not be `NULL`.
 */
int fmpi_futhark_check_error(const struct fmpi_futhark_ctx * ctx, const char * func);
/*------------------------------------------------------------------------------
    fmpi_futhark_new_data_sync()
------------------------------------------------------------------------------*/
void * fmpi_futhark_new_data_sync(
    const struct fmpi_futhark_ctx * ctx, const void * data,
    enum fmpi_type_base type, size_t dim_cnt, size_t x, size_t y, size_t z
);
/*------------------------------------------------------------------------------
    fmpi_futhark_new_data_async()
------------------------------------------------------------------------------*/
void * fmpi_futhark_new_data_async(
    const struct fmpi_futhark_ctx * ctx, const void * data,
    enum fmpi_type_base type, size_t dim_cnt, size_t x, size_t y, size_t z
);
/*------------------------------------------------------------------------------
    fmpi_futhark_free_data_sync()
------------------------------------------------------------------------------*/
int fmpi_futhark_free_data_sync(
    const struct fmpi_futhark_ctx * ctx, void * data, enum fmpi_type_base type,
    size_t dim_cnt
);
/*------------------------------------------------------------------------------
    fmpi_futhark_free_data_async()
------------------------------------------------------------------------------*/
int fmpi_futhark_free_data_async(
    const struct fmpi_futhark_ctx * ctx, void * data, enum fmpi_type_base type,
    size_t dim_cnt
);
/*------------------------------------------------------------------------------
    fmpi_futhark_get_data_sync()
------------------------------------------------------------------------------*/
void * fmpi_futhark_get_data_sync(
    const struct fmpi_futhark_ctx * ctx, void * in, void * out,
    enum fmpi_type_base type, size_t dim_cnt
);
/*------------------------------------------------------------------------------
    fmpi_futhark_get_data_async()
------------------------------------------------------------------------------*/
void * fmpi_futhark_get_data_async(
    const struct fmpi_futhark_ctx * ctx, void * in, void * out,
    enum fmpi_type_base type, size_t dim_cnt
);
/*==============================================================================
   MACRO
==============================================================================*/
/*------------------------------------------------------------------------------
    fmpi_futhark_new_1d()
------------------------------------------------------------------------------*/
/**
 * TODO
 */
#define fmpi_futhark_new_1d(ctx, array, x) \
    FMPI_GENERIC_FUNC(array, new_1d, FMPI_FUTHARK_TYPES)(ctx, array, x, 0, 0)

#define fmpi_futhark_new_2d(ctx, array, x, y) \
    FMPI_GENERIC_FUNC(array, new_2d, FMPI_FUTHARK_TYPES)(ctx, array, x, y, 0)

#define fmpi_futhark_new_3d(ctx, array, x, y, z) \
    FMPI_GENERIC_FUNC(array, new_2d, FMPI_FUTHARK_TYPES)(ctx, array, x, y, z)
/*------------------------------------------------------------------------------
    fmpi_futhark_free_1d()
------------------------------------------------------------------------------*/
/**
 * TODO
 */
#define fmpi_futhark_free_1d(ctx, array, length) \
    FMPI_GENERIC_FUNC(array, free_1d, FMPI_FUTHARK_TYPES)(ctx, array)
/*==============================================================================
    GUARD
==============================================================================*/
#endif // FMPI_FUTHARK_H_20211219234207
