// SPDX-License-Identifier: 0BSD
/*!
 * @file
 * @date 20.06.2022
 * @license{
 * BSD Zero Clause License
 *
 * Copyright (c) 2022 by Raphael Bach
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
 * REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
 * INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
 * OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 * }
 */
/*==============================================================================
    GUARD
==============================================================================*/
#ifndef FMPI_TYPE_H_20220620234502
#define FMPI_TYPE_H_20220620234502
/*==============================================================================
    INCLUDE
==============================================================================*/
// C Standard Library
#include <stddef.h> // size_t
// Internal
#include "internal/generic/fmpi_type_generic.h"
/*==============================================================================
    TYPE
==============================================================================*/
typedef enum fmpi_type_base {
    FMPI_TYPE_START = 0,
    FMPI_TYPE_i8    = 0,
    FMPI_TYPE_i16   = 1,
    FMPI_TYPE_i32   = 2,
    FMPI_TYPE_i64   = 3,
    FMPI_TYPE_u8    = 4,
    FMPI_TYPE_u16   = 5,
    FMPI_TYPE_u32   = 6,
    FMPI_TYPE_u64   = 7,
    FMPI_TYPE_f32   = 8,
    FMPI_TYPE_f64   = 9,
    FMPI_TYPE_bool  = 10,
    FMI_TYPE_END
} fmpi_type_base;
/**
 * Derived types
 *
 * - The Futhark C API passes parameters to a Futhark function differently
 * depending on their type being an `arithmetic type`, a `struct type` or an
 * `array type`.
 * - Loosely based on the definition of derived types in C17 6.2.5
 */
typedef enum fmpi_type_derived {
    FMPI_TYPE_NONE, //!< Arithmetic type
    FMPI_TYPE_ARRAY //!< Array type
} fmpi_type_derived;

typedef struct fmpi_type {
    enum fmpi_type_base base;
    enum fmpi_type_derived derived;
    size_t size;
} fmpi_type;
/*==============================================================================
    GUARD
==============================================================================*/
#endif // FMPI_TYPE_H_20220620234502
