// SPDX-License-Identifier: 0BSD
/*!
 * @file
 * @license{
 * BSD Zero Clause License
 *
 * Copyright (c) 2022 by Raphael Bach
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
 * REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
 * INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
 * OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 * }
 */
/*==============================================================================
    GUARD
==============================================================================*/
#ifndef FMPI_TASK_GENERIC_H_20220515003504
#define FMPI_TASK_GENERIC_H_20220515003504
/*==============================================================================
    INCLUDE
==============================================================================*/
// C Standard Library
#include <assert.h>
#include <stddef.h> // NULL
// External
#include "../../../external/cpl/cpl_arg.h"
#include "../../../external/cpl/cpl_map.h"
#include "../../../external/cpl/cpl_token.h"
// Internal
#include "fmpi_type_generic.h"
/*==============================================================================
    MACRO
==============================================================================*/
#define FMPI_PRIV_TASK_CONCAT(A, B)  FMPI_PRIV_TASK_CONCAT_(A, B)
#define FMPI_PRIV_TASK_CONCAT_(A, B) A##B

#define FMPI_TASK_DEFINITION(FUNC, N) \
int FUNC##_##N(const struct fmpi_ctx * ctx, const struct fmpi_task_args * args); \
int FUNC##_##N( \
    const struct fmpi_ctx * const ctx, const struct fmpi_task_args * const args \
){ \
_Pragma("GCC diagnostic ignored \"-Wincompatible-pointer-types\"")\
_Pragma("GCC diagnostic ignored \"-Wcast-qual\"")\
    assert(ctx != NULL); \
    assert(args != NULL); \
    assert(args->cnt == (N)); \
    if(args->out.type.derived == FMPI_TYPE_ARRAY) { \
        if(args->out.dim_cnt == 1) { \
            CPL_MAP_FIXED(FMPI_PRIV_TASK_RET_FUNC, CPL_EMPTY, \
                (FUNC, N, 1, ctx->fut->ctx, args->out.gpu, args->in), FMPI_TYPE_REAL) \
        } \
        if(args->out.dim_cnt == 2) { \
            CPL_MAP_FIXED(FMPI_PRIV_TASK_RET_FUNC, CPL_EMPTY, \
                (FUNC, N, 2, ctx->fut->ctx, args->out.gpu, args->in), FMPI_TYPE_REAL) \
        } \
        if(args->out.dim_cnt == 3) { \
            CPL_MAP_FIXED(FMPI_PRIV_TASK_RET_FUNC, CPL_EMPTY, \
                (FUNC, N, 3, ctx->fut->ctx, args->out.gpu, args->in), FMPI_TYPE_REAL) \
        } \
    } \
    return futhark_entry_##FUNC(ctx->fut->ctx, args->out.raw, FMPI_PRIV_TASK_ARGS_##N(args->in)); \
_Pragma("GCC diagnostic warning \"-Wcast-qual\"")\
_Pragma("GCC diagnostic warning \"-Wincompatible-pointer-types\"")\
}

#define FMPI_PRIV_TASK_RET_FUNC(FUNC, N, D, ctx, args_out, args_in, T) \
    if(args->out.type.base == FMPI_TYPE_##T) { \
        struct futhark_##T##_##D##d ** out = (struct futhark_##T##_##D##d **)&(args_out); \
        return futhark_entry_##FUNC(ctx, out, FMPI_PRIV_TASK_ARGS_##N(args_in));  \
    }

#define FMPI_TASK_REGISTER_SYNC_IMPL(FUNC, ctx, stencil, ...) \
    FMPI_PRIV_TASK_CONCAT(FMPI_PRIV_TASK_REGISTER_, CPL_ARG_COUNT(__VA_ARGS__)) \
    (FUNC, ctx, FMPI_TASK_TYPE_SYNC, stencil ,__VA_ARGS__)

#define FMPI_TASK_REGISTER_ASYNC_IMPL(FUNC, ctx, stencil, ...) \
    FMPI_PRIV_TASK_CONCAT(FMPI_PRIV_TASK_REGISTER_, CPL_ARG_COUNT(__VA_ARGS__)) \
    (FUNC, ctx, FMPI_TASK_TYPE_ASYNC, stencil ,__VA_ARGS__)

#define FMPI_PRIV_TASK_REGISTER_1(FUNC, ctx, type, stencil, arg_out) \
    fmpi_task_register((ctx), FUNC##_0, #FUNC, (type), (stencil), &(struct fmpi_task_args){ \
        .out = (arg_out), \
        .cnt = 0 \
    })

#define FMPI_PRIV_TASK_REGISTER_N(N, FUNC, ctx, type, stencil, arg_out, ...) \
    fmpi_task_register((ctx), FUNC##_##N, #FUNC, (type), (stencil), &(struct fmpi_task_args){ \
        .in = {__VA_ARGS__}, \
        .out = (arg_out), \
        .cnt = N \
    })

#define FMPI_PRIV_TASK_ARGS_1(args_in) (args_in)[0].gpu
#define FMPI_PRIV_TASK_ARGS_2(args_in) FMPI_PRIV_TASK_ARGS_1(args_in), (args_in)[1].gpu
#define FMPI_PRIV_TASK_ARGS_3(args_in) FMPI_PRIV_TASK_ARGS_2(args_in), (args_in)[2].gpu
#define FMPI_PRIV_TASK_ARGS_4(args_in) FMPI_PRIV_TASK_ARGS_3(args_in), (args_in)[3].gpu
#define FMPI_PRIV_TASK_ARGS_5(args_in) FMPI_PRIV_TASK_ARGS_4(args_in), (args_in)[4].gpu
#define FMPI_PRIV_TASK_ARGS_6(args_in) FMPI_PRIV_TASK_ARGS_5(args_in), (args_in)[5].gpu
#define FMPI_PRIV_TASK_ARGS_7(args_in) FMPI_PRIV_TASK_ARGS_6(args_in), (args_in)[6].gpu
#define FMPI_PRIV_TASK_ARGS_8(args_in) FMPI_PRIV_TASK_ARGS_7(args_in), (args_in)[7].gpu
#define FMPI_PRIV_TASK_ARGS_9(args_in) FMPI_PRIV_TASK_ARGS_8(args_in), (args_in)[8].gpu
#define FMPI_PRIV_TASK_ARGS_10(args_in) FMPI_PRIV_TASK_ARGS_9(args_in), (args_in)[9].gpu
#define FMPI_PRIV_TASK_ARGS_11(args_in) FMPI_PRIV_TASK_ARGS_10(args_in), (args_in)[10].gpu
#define FMPI_PRIV_TASK_ARGS_12(args_in) FMPI_PRIV_TASK_ARGS_11(args_in), (args_in)[11].gpu
#define FMPI_PRIV_TASK_ARGS_13(args_in) FMPI_PRIV_TASK_ARGS_12(args_in), (args_in)[12].gpu
#define FMPI_PRIV_TASK_ARGS_14(args_in) FMPI_PRIV_TASK_ARGS_13(args_in), (args_in)[13].gpu
#define FMPI_PRIV_TASK_ARGS_15(args_in) FMPI_PRIV_TASK_ARGS_14(args_in), (args_in)[14].gpu
#define FMPI_PRIV_TASK_ARGS_16(args_in) FMPI_PRIV_TASK_ARGS_15(args_in), (args_in)[15].gpu
#define FMPI_PRIV_TASK_ARGS_17(args_in) FMPI_PRIV_TASK_ARGS_16(args_in), (args_in)[16].gpu
#define FMPI_PRIV_TASK_ARGS_18(args_in) FMPI_PRIV_TASK_ARGS_17(args_in), (args_in)[17].gpu
#define FMPI_PRIV_TASK_ARGS_19(args_in) FMPI_PRIV_TASK_ARGS_18(args_in), (args_in)[18].gpu
#define FMPI_PRIV_TASK_ARGS_20(args_in) FMPI_PRIV_TASK_ARGS_19(args_in), (args_in)[19].gpu
#define FMPI_PRIV_TASK_ARGS_21(args_in) FMPI_PRIV_TASK_ARGS_20(args_in), (args_in)[20].gpu
#define FMPI_PRIV_TASK_ARGS_22(args_in) FMPI_PRIV_TASK_ARGS_21(args_in), (args_in)[21].gpu
#define FMPI_PRIV_TASK_ARGS_23(args_in) FMPI_PRIV_TASK_ARGS_22(args_in), (args_in)[22].gpu
#define FMPI_PRIV_TASK_ARGS_24(args_in) FMPI_PRIV_TASK_ARGS_23(args_in), (args_in)[23].gpu
#define FMPI_PRIV_TASK_ARGS_25(args_in) FMPI_PRIV_TASK_ARGS_24(args_in), (args_in)[24].gpu
#define FMPI_PRIV_TASK_ARGS_26(args_in) FMPI_PRIV_TASK_ARGS_25(args_in), (args_in)[25].gpu
#define FMPI_PRIV_TASK_ARGS_27(args_in) FMPI_PRIV_TASK_ARGS_26(args_in), (args_in)[26].gpu
#define FMPI_PRIV_TASK_ARGS_28(args_in) FMPI_PRIV_TASK_ARGS_27(args_in), (args_in)[27].gpu
#define FMPI_PRIV_TASK_ARGS_29(args_in) FMPI_PRIV_TASK_ARGS_28(args_in), (args_in)[28].gpu
#define FMPI_PRIV_TASK_ARGS_30(args_in) FMPI_PRIV_TASK_ARGS_29(args_in), (args_in)[29].gpu
#define FMPI_PRIV_TASK_ARGS_31(args_in) FMPI_PRIV_TASK_ARGS_30(args_in), (args_in)[30].gpu
#define FMPI_PRIV_TASK_ARGS_32(args_in) FMPI_PRIV_TASK_ARGS_31(args_in), (args_in)[31].gpu

#define FMPI_PRIV_TASK_REGISTER_2(...)  FMPI_PRIV_TASK_REGISTER_N(1, __VA_ARGS__)
#define FMPI_PRIV_TASK_REGISTER_3(...)  FMPI_PRIV_TASK_REGISTER_N(2, __VA_ARGS__)
#define FMPI_PRIV_TASK_REGISTER_4(...)  FMPI_PRIV_TASK_REGISTER_N(3, __VA_ARGS__)
#define FMPI_PRIV_TASK_REGISTER_5(...)  FMPI_PRIV_TASK_REGISTER_N(4, __VA_ARGS__)
#define FMPI_PRIV_TASK_REGISTER_6(...)  FMPI_PRIV_TASK_REGISTER_N(5, __VA_ARGS__)
#define FMPI_PRIV_TASK_REGISTER_7(...)  FMPI_PRIV_TASK_REGISTER_N(6, __VA_ARGS__)
#define FMPI_PRIV_TASK_REGISTER_8(...)  FMPI_PRIV_TASK_REGISTER_N(7, __VA_ARGS__)
#define FMPI_PRIV_TASK_REGISTER_9(...)  FMPI_PRIV_TASK_REGISTER_N(8, __VA_ARGS__)
#define FMPI_PRIV_TASK_REGISTER_10(...) FMPI_PRIV_TASK_REGISTER_N(9, __VA_ARGS__)
#define FMPI_PRIV_TASK_REGISTER_11(...) FMPI_PRIV_TASK_REGISTER_N(10, __VA_ARGS__)
#define FMPI_PRIV_TASK_REGISTER_12(...) FMPI_PRIV_TASK_REGISTER_N(11, __VA_ARGS__)
#define FMPI_PRIV_TASK_REGISTER_13(...) FMPI_PRIV_TASK_REGISTER_N(12, __VA_ARGS__)
#define FMPI_PRIV_TASK_REGISTER_14(...) FMPI_PRIV_TASK_REGISTER_N(13, __VA_ARGS__)
#define FMPI_PRIV_TASK_REGISTER_15(...) FMPI_PRIV_TASK_REGISTER_N(14, __VA_ARGS__)
#define FMPI_PRIV_TASK_REGISTER_16(...) FMPI_PRIV_TASK_REGISTER_N(15, __VA_ARGS__)
#define FMPI_PRIV_TASK_REGISTER_17(...) FMPI_PRIV_TASK_REGISTER_N(16, __VA_ARGS__)
#define FMPI_PRIV_TASK_REGISTER_18(...) FMPI_PRIV_TASK_REGISTER_N(17, __VA_ARGS__)
#define FMPI_PRIV_TASK_REGISTER_19(...) FMPI_PRIV_TASK_REGISTER_N(18, __VA_ARGS__)
#define FMPI_PRIV_TASK_REGISTER_20(...) FMPI_PRIV_TASK_REGISTER_N(19, __VA_ARGS__)
#define FMPI_PRIV_TASK_REGISTER_21(...) FMPI_PRIV_TASK_REGISTER_N(20, __VA_ARGS__)
#define FMPI_PRIV_TASK_REGISTER_22(...) FMPI_PRIV_TASK_REGISTER_N(21, __VA_ARGS__)
#define FMPI_PRIV_TASK_REGISTER_23(...) FMPI_PRIV_TASK_REGISTER_N(22, __VA_ARGS__)
#define FMPI_PRIV_TASK_REGISTER_24(...) FMPI_PRIV_TASK_REGISTER_N(23, __VA_ARGS__)
#define FMPI_PRIV_TASK_REGISTER_25(...) FMPI_PRIV_TASK_REGISTER_N(24, __VA_ARGS__)
#define FMPI_PRIV_TASK_REGISTER_26(...) FMPI_PRIV_TASK_REGISTER_N(25, __VA_ARGS__)
#define FMPI_PRIV_TASK_REGISTER_27(...) FMPI_PRIV_TASK_REGISTER_N(26, __VA_ARGS__)
#define FMPI_PRIV_TASK_REGISTER_28(...) FMPI_PRIV_TASK_REGISTER_N(27, __VA_ARGS__)
#define FMPI_PRIV_TASK_REGISTER_29(...) FMPI_PRIV_TASK_REGISTER_N(28, __VA_ARGS__)
#define FMPI_PRIV_TASK_REGISTER_30(...) FMPI_PRIV_TASK_REGISTER_N(29, __VA_ARGS__)
#define FMPI_PRIV_TASK_REGISTER_31(...) FMPI_PRIV_TASK_REGISTER_N(30, __VA_ARGS__)
#define FMPI_PRIV_TASK_REGISTER_32(...) FMPI_PRIV_TASK_REGISTER_N(31, __VA_ARGS__)
/*==============================================================================
    GUARD
==============================================================================*/
#endif // FMPI_PRIV_TASK_GENERIC_H_20220515003504
