// SPDX-License-Identifier: 0BSD
/*!
 * @file
 * @license{
 * BSD Zero Clause License
 *
 * Copyright (c) 2022 by Raphael Bach
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
 * REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
 * INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
 * OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 * }
 */
/*==============================================================================
    GUARD
==============================================================================*/
#ifndef FMPI_TYPE_INTERNAL_H_20220513200128
#define FMPI_TYPE_INTERNAL_H_20220513200128
/*==============================================================================
    INCLUDE
==============================================================================*/
// C Standard Library
#include <stdbool.h>
#include <stdint.h>
/*==============================================================================
    TYPE
==============================================================================*/
// Match C type name with Futhark type name
typedef int8_t   i8;
typedef int16_t  i16;
typedef int32_t  i32;
typedef int64_t  i64;
typedef uint8_t  u8;
typedef uint16_t u16;
typedef uint32_t u32;
typedef uint64_t u64;
typedef float    f32;
typedef double   f64;
/*==============================================================================
    DEFINE
==============================================================================*/
#define FMPI_TYPE_SIGNED \
    i8, \
    i16, \
    i32, \
    i64

#define FMPI_TYPE_UNSIGNED \
    u8, \
    u16, \
    u32, \
    u64

#define FMPI_TYPE_INTEGER \
    FMPI_TYPE_SIGNED, \
    FMPI_TYPE_UNSIGNED

#define FMPI_TYPE_FLOATING \
    f32, \
    f64

#define FMPI_TYPE_REAL \
    FMPI_TYPE_INTEGER, \
    FMPI_TYPE_FLOATING

#define FMPI_TYPE_BOOLEAN \
    bool

#define FMPI_TYPE_ALL \
    FMPI_TYPE_REAL, \
    FMPI_TYPE_BOOLEAN

#define FMPI_TYPE_DEFAULT_i8   0
#define FMPI_TYPE_DEFAULT_i16  0
#define FMPI_TYPE_DEFAULT_i32  0
#define FMPI_TYPE_DEFAULT_i64  0
#define FMPI_TYPE_DEFAULT_u8   0
#define FMPI_TYPE_DEFAULT_u16  0
#define FMPI_TYPE_DEFAULT_u32  0
#define FMPI_TYPE_DEFAULT_u64  0
#define FMPI_TYPE_DEFAULT_f32  0.0f
#define FMPI_TYPE_DEFAULT_f64  0.0
#define FMPI_TYPE_DEFAULT_bool false

// C to enum
#define FMPI_TYPE_int8_t   FMPI_TYPE_i8
#define FMPI_TYPE_int16_t  FMPI_TYPE_i16
#define FMPI_TYPE_int32_t  FMPI_TYPE_i32
#define FMPI_TYPE_int64_t  FMPI_TYPE_i64
#define FMPI_TYPE_uint8_t  FMPI_TYPE_u8
#define FMPI_TYPE_uint16_t FMPI_TYPE_u16
#define FMPI_TYPE_uint32_t FMPI_TYPE_u32
#define FMPI_TYPE_uint64_t FMPI_TYPE_u64
#define FMPI_TYPE_float    FMPI_TYPE_f32
#define FMPI_TYPE_double   FMPI_TYPE_f64
#define FMPI_TYPE_bool     FMPI_TYPE_bool
#define FMPI_TYPE__Bool    FMPI_TYPE_bool

// Futhark to MPI
#define FMPI_TYPE_MPI_i8   MPI_INT8_T
#define FMPI_TYPE_MPI_i16  MPI_INT16_T
#define FMPI_TYPE_MPI_i32  MPI_INT32_T
#define FMPI_TYPE_MPI_i64  MPI_INT64_T
#define FMPI_TYPE_MPI_u8   MPI_UINT8_T
#define FMPI_TYPE_MPI_u16  MPI_UINT16_T
#define FMPI_TYPE_MPI_u32  MPI_UINT32_T
#define FMPI_TYPE_MPI_u64  MPI_UINT64_T
#define FMPI_TYPE_MPI_f32  MPI_FLOAT
#define FMPI_TYPE_MPI_f64  MPI_DOUBLE
#define FMPI_TYPE_MPI_bool MPI_C_BOOL

// C to MPI
#define FMPI_TYPE_MPI_int8_t   MPI_INT8_T
#define FMPI_TYPE_MPI_int16_t  MPI_INT16_T
#define FMPI_TYPE_MPI_int32_t  MPI_INT32_T
#define FMPI_TYPE_MPI_int64_t  MPI_INT64_T
#define FMPI_TYPE_MPI_uint8_t  MPI_UINT8_T
#define FMPI_TYPE_MPI_uint16_t MPI_UINT16_T
#define FMPI_TYPE_MPI_uint32_t MPI_UINT32_T
#define FMPI_TYPE_MPI_uint64_t MPI_UINT64_T
#define FMPI_TYPE_MPI_float    MPI_FLOAT
#define FMPI_TYPE_MPI_double   MPI_DOUBLE
#define FMPI_TYPE_MPI_bool     MPI_C_BOOL
#define FMPI_TYPE_MPI__Bool    MPI_C_BOOL
/*==============================================================================
    GUARD
==============================================================================*/
#endif // FMPI_TYPE_INTERNAL_H_20220513200128
